
import gc
import itertools
import functools

from datetime import date
from timeit import default_timer

from ..conditional import is_tty_console
from ..conditional import is_utf_console
from ..conditional import ultrajson_installed

if ultrajson_installed:
    import ujson as json
else:
    import json

if is_utf_console:
    __vengeance_prefix__ = '\tν: '    # nu: chr(957)
else:
    __vengeance_prefix__ = '\tv: '    # ascii


def print_runtime(f):

    def runtime_wrapper(_f_):
        @functools.wraps(_f_)
        def functools_wrapper(*args, **kwargs):
            tic  = default_timer()
            retv = _f_(*args, **kwargs)
            toc  = default_timer()
            elapsed = -(tic - toc)

            s = '@{}: {}'.format(function_name(_f_), format_seconds(elapsed))
            s = vengeance_message(s)
            print(s)

            return retv

        return functools_wrapper

    return runtime_wrapper(f)


def print_performance(f=None, *, repeat=3):

    def performance_wrapper(_f_):
        @functools.wraps(_f_)
        def functools_wrapper(*args, **kwargs):
            retv  = None
            best  = None
            worst = None
            total = 0.0

            functools_repeat = functools.partial(itertools.repeat, None)
            was_gc_enabled   = gc.isenabled()
            gc.disable()

            for _ in functools_repeat(repeat):
                tic  = default_timer() * 1000
                retv = _f_(*args, **kwargs)
                toc  = default_timer() * 1000

                elapsed = -(tic - toc)
                total  += elapsed

                # best  = elapsed if (best is None)  else min(best, elapsed)
                # worst = elapsed if (worst is None) else max(worst, elapsed)

                if best is None:
                    best = elapsed
                else:
                    best = min(best, elapsed)

                if worst is None:
                    worst = elapsed
                else:
                    worst = max(worst, elapsed)

            s = ('@{}'
                 '\n        best:    {}'
                 '\n        average: {}'
                 '\n        worst:   {}'
                 .format(function_name(_f_),
                         format_milliseconds(best),
                         format_milliseconds(total / repeat),
                         format_milliseconds(worst)))
            s = vengeance_message(s)
            print(s)

            if was_gc_enabled:
                gc.enable()

            return retv

        return functools_wrapper

    if f is None:
        return performance_wrapper
    else:
        return performance_wrapper(f)


def styled(message,
           color='dark magenta',
           effect='bold'):

    # region {escape codes}
    effect_codes = {'bold':      '\x1b[1m',
                    'italic':    '\x1b[3m',
                    'underline': '\x1b[4m',
                    'end':       '\x1b[0m',
                    '':          '',
                    None:        ''}
    color_codes = {'grey':           '\x1b[29m',
                   'white':          '\x1b[30m',
                   'red':            '\x1b[31m',
                   'orange':         '\x1b[32m',
                   'yellow':         '\x1b[33m',
                   'blue':           '\x1b[34m',
                   'magenta':        '\x1b[35m',
                   'green':          '\x1b[36m',
                   'bronze':         '\x1b[37m',
                   'bright red':     '\x1b[91m',
                   'bright yellow':  '\x1b[93m',
                   'bright magenta': '\x1b[95m',
                   'bright cyan':    '\x1b[96m',
                   'dark magenta':   '\x1b[38;2;170;50;130m',
                   '':               '',
                   None:             ''}
    # endregion

    if color not in color_codes:
        raise KeyError('invalid color: {}'.format(color))
    if effect not in effect_codes:
        raise KeyError('invalid style: {}'.format(effect))

    if is_tty_console:          # TTY console doesn't support ascii escapes
        return message

    styled_message = ('{ascii_color}{effect_start}{message}{effect_end}'
                       .format(ascii_color=color_codes[color],
                               effect_start=effect_codes[effect],
                               message=message,
                               effect_end=effect_codes['end']))
    return styled_message


def vengeance_message(message):
    return __vengeance_prefix__ + message


def deprecated(message='deprecated'):

    def deprecated_wrapper(_f_):
        @functools.wraps(_f_)
        def functools_wrapper(*args, **kwargs):
            _message_ = "@{}: '{}'".format(function_name(_f_), message)
            vengeance_warning(_message_,
                              DeprecationWarning,
                              stacklevel=3,
                              stackframe=None)

            return _f_(*args, **kwargs)

        return functools_wrapper

    return deprecated_wrapper


def vengeance_warning(message,
                      category=Warning,
                      stacklevel=2,
                      stackframe=None):
    import warnings

    # region {closure functions}
    if isinstance(stacklevel, int) and stackframe is None:
        import inspect
        import traceback
        stackframe = traceback.extract_stack(inspect.currentframe(), limit=stacklevel)[0]

    def vengeance_formatwarning(*_, **__):
        nonlocal stackframe
        nonlocal category
        nonlocal message

        aligned_indent = len(__vengeance_prefix__.replace('\t', ''))
        aligned_indent = ' ' * aligned_indent
        aligned_indent = '\t' + aligned_indent

        w_category = object_name(category)
        w_message  = message.replace('\n', '\n' + aligned_indent)

        line_1 = '<{}> {}'.format(w_category, w_message)
        if stackframe is None:
            line_2 = 'File {unknown}, line {unknown}'
        else:
            line_2 = 'File "{}", line {}'.format(stackframe.filename, stackframe.lineno)

        line_1 = vengeance_message(line_1)
        line_2 = aligned_indent + line_2
        w_message = '{}\n{}\n'.format(line_1, line_2)

        # w_message = styled(w_message, 'yellow', 'bold')

        return w_message
    # endregion

    original_formatwarning = warnings.formatwarning

    warnings.formatwarning = vengeance_formatwarning
    warnings.warn(message)
    warnings.formatwarning = original_formatwarning

    # print(end='')
    # import sys
    # sys.stdout.flush()
    # sys.stderr.flush()


def format_vengeance_header(n):
    return '⟪{}⟫'.format(n)


def format_seconds(secs):
    return format_milliseconds(secs * 1000)


def format_milliseconds(ms):
    ns = ms * 1000000
    us = ms * 1000
    s  = ms / 1000

    m, s = divmod(s, 60)
    h, m = divmod(m, 60)
    d, h = divmod(h, 24)

    if d >= 1:
        f_ms = '{:.0f}d {:.0f}h {:.0f}m'.format(d, h, m)
    elif h >= 1:
        f_ms = '{:.0f}h {:.0f}m {:.0f}s'.format(h, m, s)
    elif m >= 1:
        f_ms = '{:.0f}m {:.0f}s'.format(m, s)
    elif s >= 1:
        f_ms = '{:.2f} seconds'.format(s)
    elif ms >= 1:
        f_ms = '{:.1f} ms'.format(ms)
    elif us >= 1:
        if is_utf_console:
            f_ms = '{:.0f} μs'.format(us)
        else:
            f_ms = '{:.0f} mi s'.format(us)
    else:
        f_ms = '{:.0f} ns'.format(ns)

    return f_ms


def function_name(f):
    name = f.__qualname__
    if '.' in name:
        return name

    modulename = f.__module__.split('.')[-1]
    return '{}.{}'.format(modulename, name)


def object_name(o):
    try:
        return o.__name__
    except AttributeError:
        return o.__class__.__name__


# def json_dumps_extended(o, indent=4, ensure_ascii=False):
def json_dumps_extended(o, **kwargs):
    kwargs['ensure_ascii'] = kwargs.get('ensure_ascii', False)
    kwargs['indent']       = kwargs.get('indent', 4)
    kwargs['default']      = kwargs.get('default', json_unhandled_conversion)

    if ultrajson_installed:
        del kwargs['default']

    return json.dumps(o, **kwargs)


def json_unhandled_conversion(v):
    """
    convert certain python objects to json string representations, eg:
        date, datetime, set
    """
    if isinstance(v, date):
        return v.isoformat()
    if isinstance(v, set):
        return list(v)

    raise TypeError("Object of type '{}' is not JSON serializable".format(object_name(v)))





import re
# noinspection PyUnresolvedReferences
from pythoncom import com_error as pythoncom_error

from ...conditional import ordereddict

from .lev_row_cls import lev_row_cls

from .. import excel_address
from .. import worksheet
from ..excel_constants import *

from ...util.iter import iterator_to_list
from ...util.iter import map_to_numeric_indices
from ...util.iter import modify_iteration_depth
from ...util.text import object_name


class excel_levity_cls:
    allow_focus = False

    def __init__(self, ws, *,
                       first_c=None,
                       last_c=None,
                       meta_r=0,
                       header_r=0,
                       first_r=0,
                       last_r=0):
        """
        self.is_worksheet_type, ie
            a chart that has been moved to its own worksheet
        """

        if (not isinstance(meta_r, int) or
             not isinstance(header_r, int) or
             not isinstance(first_r, int) or
             not isinstance(last_r, int)):

            raise TypeError('rows must be an integer')

        self.ws = ws
        if hasattr(ws, 'Name'):
            self.ws_name = ws.Name
        else:
            self.ws_name = ''

        self.is_worksheet_type = worksheet.is_worksheet_instance(ws)

        self.headers   = ordereddict()
        self.m_headers = ordereddict()

        self._named_ranges  = {}
        self._fixed_columns = first_c, last_c
        self._fixed_rows    = first_r, last_r

        self.first_c  = first_c
        self.last_c   = last_c
        self.meta_r   = meta_r
        self.header_r = header_r
        self.first_r  = first_r
        self.last_r   = last_r

        self.set_range_boundaries(index_meta=True,
                                  index_header=True)

    @staticmethod
    def col_letter_offset(col_str, offset):
        return excel_address.col_letter_offset(col_str, offset)

    @staticmethod
    def col_letter(col_int):
        return excel_address.col_letter(col_int)

    @staticmethod
    def col_number(col_str):
        return excel_address.col_number(col_str)

    @property
    def application(self):
        return self.ws.Application

    @property
    def workbook(self):
        return self.ws.Parent

    @property
    def named_ranges(self):
        if not self._named_ranges:
            self._named_ranges = _named_ranges_in_workbook(self.workbook)

        return self._named_ranges

    @property
    def worksheet(self):
        return self.ws

    @property
    def worksheet_name(self):
        return self.ws_name

    @property
    def meta_headers(self):
        return self.m_headers

    @property
    def header_names(self):
        return list(self.headers.keys())

    @property
    def meta_header_names(self):
        return list(self.m_headers.keys())

    @property
    def has_headers(self):
        if self.is_empty():
            return False

        return bool(self.headers) or bool(self.m_headers)

    @property
    def has_filter(self):
        return bool(self.ws.AutoFilter)

    @property
    def first_empty_row(self):
        if self.is_empty():
            return self.header_r or self.meta_r or 1

        a = '{}{}:{}{}'.format(self.first_c, self.first_r,
                               self.last_c, self.first_r)
        first_data_row = self.ws.Range(a)

        if worksheet.is_range_empty(first_data_row):
            r = self.first_r
        else:
            r = self.last_r + 1

        return r

    @property
    def first_empty_column(self):
        """ determines the first available empty column in sheet """
        if self.is_empty():
            c = self.first_c
        else:
            c = excel_address.col_letter_offset(self.last_c, 1)

        return c

    @property
    def num_cols(self):
        return excel_address.col_number(self.last_c) - excel_address.col_number(self.first_c) + 1

    @property
    def num_rows(self):
        return int(self.last_r) - int(self.first_r) + 1

    def is_empty(self):
        if self.last_r > self.first_r:
            return False

        r_1 = self.header_r or self.meta_r or 1
        r_2 = self.last_r
        a = '{}{}:{}{}'.format(self.first_c, r_1, self.last_c, r_2)

        return worksheet.is_range_empty(self.ws.Range(a))

    def rows(self, r_1='*h', r_2='*l'):
        if self.is_empty():
            return [[]]

        a = '*f {}:*l {}'.format(r_1, r_2)
        excel_range = self.excel_range(a)

        return worksheet.escape_excel_range_errors(excel_range)

    def lev_rows(self, r_1='*h', r_2='*l'):
        if self.is_empty():
            return [[]]

        if self.headers:
            headers = map_to_numeric_indices(self.headers.keys())
        elif self.m_headers:
            headers = map_to_numeric_indices(self.m_headers.keys())
        else:
            headers = {}

        reserved = headers.keys() & set(lev_row_cls.reserved_names())
        if reserved:
            raise NameError("reserved name(s) {} found in header row {}"
                            .format(list(reserved), list(headers.keys())))

        a = '*f {}:*l {}'.format(r_1, r_2)
        excel_range = self.excel_range(a)

        r_1 = excel_range.Row
        c_1, c_2 = self.first_c, self.last_c

        for r, row in enumerate(worksheet.escape_excel_range_errors(excel_range), r_1):
            a = '${}${}:${}${}'.format(c_1, r, c_2, r)
            yield lev_row_cls(headers, row, a)

    def activate(self):
        if self.allow_focus:
            worksheet.activate_worksheet(self.ws)

    def clear_filter(self):
        if not self.has_filter:
            return

        if worksheet.is_filtered(self.ws):
            self.set_range_boundaries()

    def remove_filter(self):
        if not self.has_filter:
            return

        self.ws.AutoFilterMode = False

    def reapply_filter(self, c_1='*f', c_2='*l'):
        if self.header_r > 0:
            r = '*h'
        else:
            r = '*f'

        a = '{} {}:{} {}'.format(c_1, r, c_2, r)
        excel_range = self.excel_range(a)

        if excel_range.Cells.Count == 1:
            a = '{} {}:{} {}'.format(c_1, r, c_2, self.last_r + 1)
            excel_range = self.excel_range(a)

        self.remove_filter()
        excel_range.AutoFilter(*(1,))

    def calculate(self):
        self.excel_range('*f *h: *l *l').Calculate()

    def clear(self, reference,
                    clear_values=True,
                    clear_colors=False):

        excel_range = self.excel_range(reference)
        _, r_1, _, r_2 = worksheet.parse_range(excel_range)

        if clear_values:
            excel_range.ClearContents()

            index_meta   = (r_1 <= self.meta_r)
            index_header = (r_1 <= self.header_r)
            self.set_range_boundaries(index_meta, index_header)

        if clear_colors:
            excel_range.Interior.Color = xlNone

    def set_range_boundaries(self, index_meta=True, index_header=True):
        """ find the edges of data in worksheet

        worksheet filter MUST be cleared from worksheet to
        determine these boundaries correctly
        """

        if not self.is_worksheet_type:
            self.first_c = ''
            self.last_c  = ''
            self.first_r = 0
            self.last_r  = 0
            return

        worksheet.clear_worksheet_filter(self.ws)
        self.__range_boundaries()

        if index_meta:
            self.__index_meta_columns()

        if index_header:
            self.__index_header_columns()

    def __range_boundaries(self):
        used_range = self.ws.UsedRange

        first_c, last_c = self._fixed_columns
        first_r, last_r = self._fixed_rows

        self.first_c = first_c or worksheet.first_col(used_range)
        self.last_c  = last_c  or worksheet.last_col(used_range, default=self.first_c)

        r_1 = max(self.meta_r, self.header_r) + 1
        r_2 = used_range.Rows.Count

        a = '{}{}:{}{}'.format(self.first_c, r_1,
                               self.last_c,  r_2)
        excel_range = self.ws.Range(a)

        self.first_r = first_r or worksheet.first_row(excel_range, default=r_1)
        self.last_r  = last_r  or worksheet.last_row(excel_range,  default=self.first_r)

        self.first_c = excel_address.col_letter(self.first_c)
        self.last_c  = excel_address.col_letter(self.last_c)
        self.first_r = int(self.first_r)
        self.last_r  = int(self.last_r)

    @classmethod
    def index_headers(cls, ws, row_int=None):
        if ws.__class__.__name__ != '_Worksheet':
            return {}

        if row_int is None:
            row_int = worksheet.first_row(ws)

        c = excel_address.col_letter(ws.UsedRange.Columns.Count)
        a = '{}{}:{}{}'.format('A', row_int, c, row_int)
        excel_range = ws.Range(a)

        return cls.__index_row_headers(excel_range)

    @classmethod
    def __index_row_headers(cls, excel_range):
        row = excel_range.Rows(1)
        row = worksheet.escape_excel_range_errors(row)[0]
        if not any(row):
            return ordereddict()

        c_1 = excel_range.Column
        headers = map_to_numeric_indices(row, c_1)
        headers = ordereddict((h, excel_address.col_letter(v)) for h, v in headers.items())

        return headers

    def __index_headers(self, row_ref):
        a = '*f {} :*l {}'.format(row_ref, row_ref)
        excel_range = self.excel_range(a)

        return self.__index_row_headers(excel_range)

    def __index_meta_columns(self):
        if self.meta_r == 0:
            return

        self.m_headers = self.__index_headers('meta_r')

    def __index_header_columns(self):
        if self.header_r == 0:
            return

        self.headers = self.__index_headers('header_r')

    def excel_range(self, reference):
        if not self.is_worksheet_type:
            ws_type = object_name(self.ws)
            raise TypeError('{} is not an Excel worksheet '.format(ws_type))

        try:
            a = self.excel_address(reference)
            excel_range = self.ws.Range(a)
        except pythoncom_error:
            excel_range = self.named_ranges.get(reference)

        if excel_range is None:
            raise ValueError("Invalid Range reference '{}'".format(reference))

        return excel_range

    def excel_address(self, reference):
        if ':' in reference:
            a_1, a_2 = reference.split(':')
            c_1, r_1 = _reference_to_col_row(self, a_1)
            c_2, r_2 = _reference_to_col_row(self, a_2)

            a = '{}{}:{}{}'.format(c_1, r_1, c_2, r_2)
        else:
            c_1, r_1 = _reference_to_col_row(self, reference)
            a = '{}{}'.format(c_1, r_1)

        return a

    def __getitem__(self, reference):
        return self.excel_range(reference)

    def __setitem__(self, reference, v):
        """ write value(s) to excel range """
        excel_range = self.excel_range(reference)
        m = self.__validate_matrix_within_range_boundaries(v, excel_range)

        worksheet.write_to_excel_range(m, excel_range)

        r_1 = excel_range.Row
        index_meta   = (r_1 <= self.meta_r)
        index_header = (r_1 <= self.header_r)
        self.set_range_boundaries(index_meta, index_header)

        if self.has_filter:
            self.reapply_filter()

    def __iter__(self):
        for row in self.lev_rows('*f'):
            yield row

    def __repr__(self):
        if not self.is_worksheet_type:
            return "{{}}: '{}'".format(self.ws.__class__.__name__, self.ws_name)

        if self.first_c and self.last_c:
            a = "{}{}:{}{}".format(self.first_c,
                                   self.header_r,
                                   self.last_c,
                                   self.last_r)
        else:
            a = '{unknown address}'

        return "'{}' {}".format(self.ws_name, a)

    def __validate_matrix_within_range_boundaries(self, v, excel_range):
        """
        if lev has fixed columns or rows, these should not be exceeded
        make sure matrix fits in allowed destination space
        """
        m = iterator_to_list(v)
        m = modify_iteration_depth(m, depth=2)

        col_max = num_cols = len(m[0])
        row_max = num_rows = len(m)

        first_c, last_c = self._fixed_columns
        first_r, last_r = self._fixed_rows

        c_1 = excel_range.Column
        r_1 = excel_range.Row

        if last_c:
            first_c = excel_address.col_number(first_c) or excel_address.col_number(c_1)
            last_c  = excel_address.col_number(last_c)
            col_max = (last_c - first_c) + 1

        if last_r:
            first_r = int(first_r) or int(r_1)
            last_r  = int(last_r)
            row_max = (last_r - first_r) + 1

        if num_cols > col_max:
            raise ValueError('Number of columns in data exceeds fixed destination range')

        if num_rows > row_max:
            raise ValueError('Number of rows in data exceeds fixed destination range')

        return m


def _named_ranges_in_workbook(wb):
    named_ranges = {}

    for nr in wb.Names:
        if nr.Visible:
            try:
                named_ranges[nr.Name] = nr.RefersToRange
            except pythoncom_error:
                continue

    return named_ranges


def _reference_to_col_row(lev, reference):
    reference = __reference_to_property_names(reference)
    col, row  = __property_names_to_value(lev, reference)

    if col is None or row is None:
        col, row = __parse_characters_from_digits(reference, col, row)

    return col, row


def __reference_to_property_names(reference):
    """
    eg:
        'header_c first_r' = __reference_to_property_names('*h *f')
    """
    anchor_names = {'*m': 'meta',
                    '*h': 'header',
                    '*f': 'first',
                    '*l': 'last'}

    anchor_re = re.compile('''
         (?P<col>^[*][fla])
        |(?P<row>[*][mhfla]$)
    ''', re.X | re.I)

    reference = reference.strip()

    for match in anchor_re.finditer(reference):
        name  = match.lastgroup
        value = match.group(0)

        if name == 'col':
            if value == '*a':
                col = 'first_empty_column '
            else:
                col = anchor_names[value] + '_c '

            reference = reference.replace(value, col, 1)

        elif name == 'row':
            if value == '*a':
                row = 'first_empty_row '
            else:
                row = ' ' + anchor_names[value] + '_r'

            reference = reference.replace(value, row, 1)

    # replace multiple spaces with single space
    reference = ' '.join(reference.split())

    return reference


def __property_names_to_value(lev, reference):
    if ' ' not in reference:
        return None, None

    # replace multiple spaces with single space
    reference = ' '.join(reference.split())

    splits = reference.split(' ')
    col = __col_row_to_value(lev, splits[0])
    row = __col_row_to_value(lev, splits[1])

    return col, row


def __col_row_to_value(lev, reference):
    if reference in lev.headers:
        literal = lev.headers[reference]
    elif reference in lev.m_headers:
        literal = lev.m_headers[reference]
    elif reference in lev.__dict__:
        literal = lev.__dict__[reference]
    elif reference in lev.__class__.__dict__:
        literal = getattr(lev, reference)
    else:
        literal = None

    return literal


def __parse_characters_from_digits(reference, col, row):

    address_re = re.compile(r'''
         (?P<col>^[$]?[a-z]{1,2})(?=[\d* ])
        |(?P<row>[$]?[\d]+$)
    ''', re.X | re.I)

    for match in address_re.finditer(reference):
        name  = match.lastgroup
        value = match.group(0)

        if (name == 'col') and (col is None):
            col = value
        elif (name == 'row') and (row is None):
            row = value

    return col, row





from datetime import date
from datetime import datetime

# noinspection PyUnresolvedReferences
from pythoncom import com_error as pythoncom_error

from .excel_constants import *
from .workbook import excel_application_to_foreground

from .excel_address import col_letter
from .excel_address import col_number
from .excel_address import max_cols as excel_max_cols
from .excel_address import max_rows as excel_max_rows

from ..util.iter import iterator_to_list
from ..util.iter import modify_iteration_depth
from ..util.iter import is_exhaustable


def get_worksheet(wb,
                  ws,
                  *,
                  clear_filter=False,
                  activate=False):

    if is_worksheet_instance(ws):
        return ws

    if wb is None:
        raise AssertionError("Excel workbook has not been set, cannot retrieve worksheet: '{}'".format(ws))

    try:
        ws = wb.Sheets[ws]
    except pythoncom_error as e:
        raise NameError("'{}' worksheet not found in '{}'".format(ws, wb.Name)) from e

    # if chart that has been moved to its own worksheet, these calls will fail
    # if ws.__class__.__name__ != '_Worksheet':

    if clear_filter:
        clear_worksheet_filter(ws)

    if activate:
        ws.Visible = True
        excel_application_to_foreground(ws.Application)
        ws.Activate()

    return ws


def is_worksheet_instance(o):
    return o.__class__.__name__ in ('CDispatch', '_Worksheet')


def first_row(excel_range, default=1):

    search = range_find(excel_range,
                        what='*',
                        look_at=xlPart,
                        search_order=xlByRows,
                        search_direction=xlNext)

    if search:
        r = max(search.Row, default)
    else:
        r = default

    return int(r)


def last_row(excel_range, default=1):

    search = range_find(excel_range,
                        what='*',
                        look_at=xlPart,
                        search_order=xlByRows,
                        search_direction=xlPrevious)

    if search:
        r = max(search.Row, default)
    else:
        r = default

    return int(r)


def first_col(excel_range, default='A'):

    search = range_find(excel_range,
                        what='*',
                        look_at=xlPart,
                        search_order=xlByColumns,
                        search_direction=xlNext)

    if search:
        c = max(search.Column, col_number(default))
    else:
        c = default

    return col_letter(c)


def last_col(excel_range, default='A'):

    search = range_find(excel_range,
                        what='*',
                        look_at=xlPart,
                        search_order=xlByColumns,
                        search_direction=xlPrevious)

    if search:
        c = max(search.Column, col_number(default))
    else:
        c = default

    return col_letter(c)


def range_find(excel_range,
               what,
               after=None,
               look_in=xlValues,
               look_at=xlWhole,
               search_order=xlByRows,
               search_direction=xlPrevious,
               match_case=False):

    if after is None:
        if search_direction == xlNext:
            after = last_cell(excel_range)
        elif search_direction == xlPrevious:
            after = first_cell(excel_range)

    search = excel_range.Find(what,
                              after,
                              look_in,
                              look_at,
                              search_order,
                              search_direction,
                              match_case)

    return search


def first_cell(excel_range):
    return excel_range.Cells(1)


def last_cell(excel_range):
    """
    excel_range.Cells(excel_range.Cells.Count) is not always reliable
    and can cause an overflow error, but when excel_range.Address
    only provides row address, something like '$1:$20', this method must be used

    # '$1:$20' = excel_range.Address
    # excel_range.Parent.Rows(a)
    """
    try:
        a = excel_range.Address.split(':')[-1]
        return excel_range.Parent.Range(a)
    except pythoncom_error:
        return excel_range.Cells(excel_range.Cells.Count)


def is_filtered(ws):
    return (ws.AutoFilter is not None) and bool(ws.AutoFilter.FilterMode)


# noinspection PyProtectedMember
def clear_worksheet_filter(ws):
    if is_filtered(ws):
        ws._AutoFilter.ShowAllData()


def is_range_empty(excel_range):
    num_blank = excel_range.Application.WorksheetFunction.CountBlank(excel_range)
    num_cells = excel_range.Cells.Count

    return num_blank == num_cells


def activate_worksheet(ws):
    excel_application_to_foreground(ws.Application)

    ws.Parent.Activate()

    ws.Visible = True
    ws.Activate()


def write_to_excel_range(v, excel_range):
    m = validate_matrix_within_max_worksheet_dimensions(v)
    m = list(convert_python_types(m))

    num_cols = len(m[0])
    num_rows = len(m)
    excel_range_dest = excel_range.Resize(num_rows, num_cols)

    excel_range.Parent.Range(excel_range, excel_range_dest).Value = m


def escape_excel_range_errors(excel_range):
    m = excel_range.Value
    m = modify_iteration_depth(m, 2)
    m = [list(row) for row in m]

    try:
        range_errors = excel_range.SpecialCells(xlCellTypeFormulas, xlErrors)
        cell = first_cell(excel_range)
        r_0  = cell.Row
        c_0  = cell.Column

        for cell_error in range_errors:
            r = cell_error.Row    - r_0
            c = cell_error.Column - c_0
            m[r][c] = excel_errors.get(m[r][c], 'unknown error')

    except pythoncom_error:
        pass

    return m


def convert_python_types(m):
    """ convert python types so they are compatible with Excel

    for each row in matrix:
        for each value in row:
            convert datetime.date to datetime.datetime (pywintypes time?)
            convert non-primitives to repr()
    """
    # region {closure function}
    # noinspection PyPep8Naming
    NoneType = type(None)

    def convert_excel_value(v):
        if isinstance(v, (NoneType, bool, int, float, str)):
            _v_ = v
        elif isinstance(v, date):
            _v_ = datetime(v.year, v.month, v.day)
        else:
            _v_ = repr(v)

        return _v_
    # endregion

    if is_exhaustable(m):
        # m = iterator_to_list(m)
        # m = modify_iteration_depth(m, depth=2)
        raise TypeError('matrix must be a list of lists, not a generator')

    max_cols = len(m[0])

    for i, row in enumerate(m):
        if (max_cols - len(row)) != 0:
            raise ValueError('cannot write to Excel worksheet, jagged columns at row {:,}\n\t'.format(i))

        yield tuple(convert_excel_value(v) for v in row)


def validate_matrix_within_max_worksheet_dimensions(v):
    """ ensure matrix fits within Excel's column and row maximum """
    m = iterator_to_list(v)
    m = modify_iteration_depth(m, depth=2)

    num_rows = len(m)
    num_cols = max(map(len, m))

    if num_rows > excel_max_rows:
        raise ValueError("number of rows in matrix ({:,}) exceeds Excel's row limit ({:,})"
                         .format(num_rows, excel_max_rows))

    if num_cols > excel_max_cols:
        raise ValueError("number of columns in matrix ({:,}) exceeds Excel's column limit({:,})"
                         "(did you mean to transpose this matrix first?)"
                         .format(num_cols, excel_max_cols))

    return m


def parse_range(excel_range):
    c_1 = first_cell(excel_range).Column
    c_1 = col_letter(c_1)

    c_2 = last_cell(excel_range).Column
    c_2 = col_letter(c_2)

    r_1 = first_cell(excel_range).Row
    r_2 = last_cell(excel_range).Row

    return c_1, r_1, c_2, r_2





import os
import setuptools
from setuptools import setup


__version__ = '1.1.9'
__release__ = '$release 47'
long_description = ('https://github.com/michael-ross-ven/vengeance/blob/master/README.md'
                    '\n\n(specialize this for pypi.org later)')


def __move_win32com_gencache_folder():
    """
    move win32com gen_py cache files from temp to site-packages folder
    helps prevent win32com EnsureDispatch() call rejection due to corrupted COM files
    """
    try:
        import shutil
        import site

        # win32com site-package installed?
        if not os.path.exists(site.getsitepackages()[1] + '\\win32com\\'):
            return

        new_gencache_folder = site.getsitepackages()[1] + '\\win32com\\gen_py\\'
        old_gencache_folder = os.environ['userprofile'] + '\\AppData\\Local\\Temp\\gen_py\\'

        if not os.path.exists(new_gencache_folder):
            os.makedirs(new_gencache_folder)
            if os.path.exists(old_gencache_folder):
                shutil.rmtree(old_gencache_folder)

    except Exception:
        pass


if __name__ == '__main__':
    setup(name='vengeance',
          version=__version__,
          description='Library focusing on row-major organization of tabular data and control over '
                      'the Excel application',
          long_description=long_description,
          url='https://github.com/michael-ross-ven/vengeance',
          author='Michael Ross',
          author_email='',
          license='MIT',
          install_requires=('comtypes',
                            'pypiwin32'),
          extra_require=('python-dateutil',
                         'ujson',
                         'numpy'),
          packages=setuptools.find_packages(),
          classifiers=[
              "Programming Language :: Python :: 3",
              "License :: OSI Approved :: MIT License"
            ]

          )

    __move_win32com_gencache_folder()





""" import vengeance as ven """

from .version import *
from .util import *
from .classes import *
from .conditional import loads_excel_module

''' 
attempt to load Excel module, but allow it to fail if it's not expected to support Windows COM
loads_excel_module = (not is_pypy_interpreter) and is_windows_os
    
    vgc.open_workbook, 
    vgc.close_workbook,
    vgc.excel_levity_cls, 
    etc 
'''
if loads_excel_module:
    from .excel_com import *
